/* @copyright Itential, LLC 2019 (pre-modifications) */

// Set globals
/* global describe it log pronghornProps */
/* eslint global-require: warn */
/* eslint no-unused-vars: warn */

// include required items for testing & logging
const assert = require('assert');
const fs = require('fs-extra');
const mocha = require('mocha');
const path = require('path');
const util = require('util');
const winston = require('winston');
const execute = require('child_process').execSync;
const { expect } = require('chai');
const { use } = require('chai');
const td = require('testdouble');

const anything = td.matchers.anything();

// stub and attemptTimeout are used throughout the code so set them here
let logLevel = 'none';
const stub = true;
const isRapidFail = false;
const attemptTimeout = 120000;

// these variables can be changed to run in integrated mode so easier to set them here
// always check these in with bogus data!!!
const host = 'replace.hostorip.here';
const username = 'username';
const password = 'password';
const protocol = 'http';
const port = 80;
const sslenable = false;
const sslinvalid = false;

// these are the adapter properties. You generally should not need to alter
// any of these after they are initially set up
global.pronghornProps = {
  pathProps: {
    encrypted: false
  },
  adapterProps: {
    adapters: [{
      id: 'Test-awscloudFormation',
      type: 'AwsCloudFormation',
      properties: {
        host,
        region: 'us-east-1',
        port,
        base_path: '/api',
        version: 'v2016-11-15',
        cache_location: 'none',
        encode_pathvars: true,
        save_metric: false,
        stub,
        protocol,
        authentication: {
          auth_method: 'no_authentication',
          username,
          password,
          token: '',
          invalid_token_error: 401,
          token_timeout: -1,
          token_cache: 'local',
          auth_field: 'header.headers.Authorization',
          auth_field_format: 'Basic {b64}{username}:{password}{/b64}',
          auth_logging: false,
          access_key: 'access',
          secret_key: 'secret'
        },
        healthcheck: {
          type: 'none',
          frequency: 60000,
          query_object: {}
        },
        throttle: {
          throttle_enabled: false,
          number_pronghorns: 1,
          sync_async: 'sync',
          max_in_queue: 1000,
          concurrent_max: 1,
          expire_timeout: 0,
          avg_runtime: 200,
          priorities: [
            {
              value: 0,
              percent: 100
            }
          ]
        },
        request: {
          number_redirects: 0,
          number_retries: 3,
          limit_retry_error: 0,
          failover_codes: [],
          attempt_timeout: attemptTimeout,
          global_request: {
            payload: {},
            uriOptions: {},
            addlHeaders: {},
            authData: {}
          },
          healthcheck_on_timeout: true,
          return_raw: true,
          archiving: false,
          return_request: false
        },
        proxy: {
          enabled: false,
          host: '',
          port: 1,
          protocol: 'http',
          username: '',
          password: ''
        },
        ssl: {
          ecdhCurve: '',
          enabled: sslenable,
          accept_invalid_cert: sslinvalid,
          ca_file: '',
          key_file: '',
          cert_file: '',
          secure_protocol: '',
          ciphers: ''
        },
        mongo: {
          host: '',
          port: 0,
          database: '',
          username: '',
          password: '',
          replSet: '',
          db_ssl: {
            enabled: sslenable,
            accept_invalid_cert: false,
            ca_file: '',
            key_file: '',
            cert_file: ''
          }
        }
      }
    }]
  }
};

global.$HOME = `${__dirname}/../..`;

// set the log levels that Pronghorn uses, spam and trace are not defaulted in so without
// this you may error on log.trace calls.
const myCustomLevels = {
  levels: {
    spam: 6,
    trace: 5,
    debug: 4,
    info: 3,
    warn: 2,
    error: 1,
    none: 0
  }
};

// need to see if there is a log level passed in
process.argv.forEach((val) => {
  // is there a log level defined to be passed in?
  if (val.indexOf('--LOG') === 0) {
    // get the desired log level
    const inputVal = val.split('=')[1];

    // validate the log level is supported, if so set it
    if (Object.hasOwnProperty.call(myCustomLevels.levels, inputVal)) {
      logLevel = inputVal;
    }
  }
});

// need to set global logging
global.log = winston.createLogger({
  level: logLevel,
  levels: myCustomLevels.levels,
  transports: [
    new winston.transports.Console()
  ]
});

/**
 * Runs the error asserts for the test
 */
function runErrorAsserts(data, error, code, origin, displayStr) {
  assert.equal(null, data);
  assert.notEqual(undefined, error);
  assert.notEqual(null, error);
  assert.notEqual(undefined, error.IAPerror);
  assert.notEqual(null, error.IAPerror);
  assert.notEqual(undefined, error.IAPerror.displayString);
  assert.notEqual(null, error.IAPerror.displayString);
  assert.equal(code, error.icode);
  assert.equal(origin, error.IAPerror.origin);
  assert.equal(displayStr, error.IAPerror.displayString);
}

// require the adapter that we are going to be using
const AwsCloudFormation = require('../../adapter.js');

// delete the .DS_Store directory in entities -- otherwise this will cause errors
const dirPath = path.join(__dirname, '../../entities/.DS_Store');
if (fs.existsSync(dirPath)) {
  try {
    fs.removeSync(dirPath);
    console.log('.DS_Store deleted');
  } catch (e) {
    console.log('Error when deleting .DS_Store:', e);
  }
}

// begin the testing - these should be pretty well defined between the describe and the it!
describe('[unit] AwsCloudFormation Adapter Test', () => {
  describe('AwsCloudFormation Class Tests', () => {
    const a = new AwsCloudFormation(
      pronghornProps.adapterProps.adapters[0].id,
      pronghornProps.adapterProps.adapters[0].properties
    );

    if (isRapidFail) {
      const state = {};
      state.passed = true;

      mocha.afterEach(function x() {
        state.passed = state.passed
        && (this.currentTest.state === 'passed');
      });
      mocha.beforeEach(function x() {
        if (!state.passed) {
          return this.currentTest.skip();
        }
        return true;
      });
    }

    describe('#class instance created', () => {
      it('should be a class with properties', (done) => {
        try {
          assert.notEqual(null, a);
          assert.notEqual(undefined, a);
          assert.notEqual(null, a.allProps);
          const check = global.pronghornProps.adapterProps.adapters[0].properties.healthcheck.type;
          assert.equal(check, a.healthcheckType);
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('adapterBase.js', () => {
      it('should have an adapterBase.js', (done) => {
        try {
          fs.exists('adapterBase.js', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    let wffunctions = [];
    describe('#getWorkflowFunctions', () => {
      it('should retrieve workflow functions', (done) => {
        try {
          wffunctions = a.getWorkflowFunctions([]);

          try {
            assert.notEqual(0, wffunctions.length);
            done();
          } catch (err) {
            log.error(`Test Failure: ${err}`);
            done(err);
          }
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('package.json', () => {
      it('should have a package.json', (done) => {
        try {
          fs.exists('package.json', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('package.json should be validated', (done) => {
        try {
          const packageDotJson = require('../../package.json');
          const { PJV } = require('package-json-validator');
          const options = {
            warnings: true, // show warnings
            recommendations: true // show recommendations
          };
          const results = PJV.validate(JSON.stringify(packageDotJson), 'npm', options);

          if (results.valid === false) {
            log.error('The package.json contains the following errors: ');
            log.error(util.inspect(results));
            assert.equal(true, results.valid);
          } else {
            assert.equal(true, results.valid);
          }

          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('package.json should be customized', (done) => {
        try {
          const packageDotJson = require('../../package.json');
          assert.notEqual(-1, packageDotJson.name.indexOf('aws_cloudformation'));
          assert.notEqual(undefined, packageDotJson.version);
          assert.notEqual(null, packageDotJson.version);
          assert.notEqual('', packageDotJson.version);
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('pronghorn.json', () => {
      it('should have a pronghorn.json', (done) => {
        try {
          fs.exists('pronghorn.json', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('pronghorn.json should be customized', (done) => {
        try {
          const pronghornDotJson = require('../../pronghorn.json');
          assert.notEqual(-1, pronghornDotJson.id.indexOf('aws_cloudformation'));
          assert.equal('AwsCloudFormation', pronghornDotJson.export);
          assert.equal('AwsCloudFormation', pronghornDotJson.title);
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('pronghorn.json should only expose workflow functions', (done) => {
        try {
          const pronghornDotJson = require('../../pronghorn.json');

          for (let m = 0; m < pronghornDotJson.methods.length; m += 1) {
            let found = false;
            let paramissue = false;

            for (let w = 0; w < wffunctions.length; w += 1) {
              if (pronghornDotJson.methods[m].name === wffunctions[w]) {
                found = true;
                const methLine = execute(`grep "  ${wffunctions[w]}(" adapter.js | grep "callback) {"`).toString();
                let wfparams = [];

                if (methLine && methLine.indexOf('(') >= 0 && methLine.indexOf(')') >= 0) {
                  const temp = methLine.substring(methLine.indexOf('(') + 1, methLine.lastIndexOf(')'));
                  wfparams = temp.split(',');

                  for (let t = 0; t < wfparams.length; t += 1) {
                    // remove default value from the parameter name
                    wfparams[t] = wfparams[t].substring(0, wfparams[t].search(/=/) > 0 ? wfparams[t].search(/#|\?|=/) : wfparams[t].length);
                    // remove spaces
                    wfparams[t] = wfparams[t].trim();

                    if (wfparams[t] === 'callback') {
                      wfparams.splice(t, 1);
                    }
                  }
                }

                // if there are inputs defined but not on the method line
                if (wfparams.length === 0 && (pronghornDotJson.methods[m].input
                    && pronghornDotJson.methods[m].input.length > 0)) {
                  paramissue = true;
                } else if (wfparams.length > 0 && (!pronghornDotJson.methods[m].input
                    || pronghornDotJson.methods[m].input.length === 0)) {
                  // if there are no inputs defined but there are on the method line
                  paramissue = true;
                } else {
                  for (let p = 0; p < pronghornDotJson.methods[m].input.length; p += 1) {
                    let pfound = false;
                    for (let wfp = 0; wfp < wfparams.length; wfp += 1) {
                      if (pronghornDotJson.methods[m].input[p].name.toUpperCase() === wfparams[wfp].toUpperCase()) {
                        pfound = true;
                      }
                    }

                    if (!pfound) {
                      paramissue = true;
                    }
                  }
                  for (let wfp = 0; wfp < wfparams.length; wfp += 1) {
                    let pfound = false;
                    for (let p = 0; p < pronghornDotJson.methods[m].input.length; p += 1) {
                      if (pronghornDotJson.methods[m].input[p].name.toUpperCase() === wfparams[wfp].toUpperCase()) {
                        pfound = true;
                      }
                    }

                    if (!pfound) {
                      paramissue = true;
                    }
                  }
                }

                break;
              }
            }

            if (!found) {
              // this is the reason to go through both loops - log which ones are not found so
              // they can be worked
              log.error(`${pronghornDotJson.methods[m].name} not found in workflow functions`);
            }
            if (paramissue) {
              // this is the reason to go through both loops - log which ones are not found so
              // they can be worked
              log.error(`${pronghornDotJson.methods[m].name} has a parameter mismatch`);
            }
            assert.equal(true, found);
            assert.equal(false, paramissue);
          }
          done();
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('pronghorn.json should expose all workflow functions', (done) => {
        try {
          const pronghornDotJson = require('../../pronghorn.json');
          for (let w = 0; w < wffunctions.length; w += 1) {
            let found = false;

            for (let m = 0; m < pronghornDotJson.methods.length; m += 1) {
              if (pronghornDotJson.methods[m].name === wffunctions[w]) {
                found = true;
                break;
              }
            }

            if (!found) {
              // this is the reason to go through both loops - log which ones are not found so
              // they can be worked
              log.error(`${wffunctions[w]} not found in pronghorn.json`);
            }
            assert.equal(true, found);
          }
          done();
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      });
    });

    describe('propertiesSchema.json', () => {
      it('should have a propertiesSchema.json', (done) => {
        try {
          fs.exists('propertiesSchema.json', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('propertiesSchema.json should be customized', (done) => {
        try {
          const propertiesDotJson = require('../../propertiesSchema.json');
          assert.equal('adapter-aws_cloudformation', propertiesDotJson.$id);
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('error.json', () => {
      it('should have an error.json', (done) => {
        try {
          fs.exists('error.json', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('sampleProperties.json', () => {
      it('should have a sampleProperties.json', (done) => {
        try {
          fs.exists('sampleProperties.json', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('#checkProperties', () => {
      it('should have a checkProperties function', (done) => {
        try {
          assert.equal(true, typeof a.checkProperties === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('the sample properties should be good - if failure change the log level', (done) => {
        try {
          const samplePropsJson = require('../../sampleProperties.json');
          const clean = a.checkProperties(samplePropsJson.properties);

          try {
            assert.notEqual(0, Object.keys(clean));
            assert.equal(undefined, clean.exception);
            assert.notEqual(undefined, clean.host);
            assert.notEqual(null, clean.host);
            assert.notEqual('', clean.host);
            done();
          } catch (err) {
            log.error(`Test Failure: ${err}`);
            done(err);
          }
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('README.md', () => {
      it('should have a README', (done) => {
        try {
          fs.exists('README.md', (val) => {
            assert.equal(true, val);
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('README.md should be customized', (done) => {
        try {
          fs.readFile('README.md', 'utf8', (err, data) => {
            assert.equal(-1, data.indexOf('[System]'));
            assert.equal(-1, data.indexOf('[system]'));
            assert.equal(-1, data.indexOf('[version]'));
            assert.equal(-1, data.indexOf('[namespace]'));
            done();
          });
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('#connect', () => {
      it('should have a connect function', (done) => {
        try {
          assert.equal(true, typeof a.connect === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('#healthCheck', () => {
      it('should have a healthCheck function', (done) => {
        try {
          assert.equal(true, typeof a.healthCheck === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
    });

    describe('#checkActionFiles', () => {
      it('should have a checkActionFiles function', (done) => {
        try {
          assert.equal(true, typeof a.checkActionFiles === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('the action files should be good - if failure change the log level as most issues are warnings', (done) => {
        try {
          const clean = a.checkActionFiles();

          try {
            for (let c = 0; c < clean.length; c += 1) {
              log.error(clean[c]);
            }
            assert.equal(0, clean.length);
            done();
          } catch (err) {
            log.error(`Test Failure: ${err}`);
            done(err);
          }
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#encryptProperty', () => {
      it('should have a encryptProperty function', (done) => {
        try {
          assert.equal(true, typeof a.encryptProperty === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      });
      it('should get base64 encoded property', (done) => {
        try {
          a.encryptProperty('testing', 'base64', (data, error) => {
            try {
              assert.equal(undefined, error);
              assert.notEqual(undefined, data);
              assert.notEqual(null, data);
              assert.notEqual(undefined, data.response);
              assert.notEqual(null, data.response);
              assert.equal(0, data.response.indexOf('{code}'));
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should get encrypted property', (done) => {
        try {
          a.encryptProperty('testing', 'encrypt', (data, error) => {
            try {
              assert.equal(undefined, error);
              assert.notEqual(undefined, data);
              assert.notEqual(null, data);
              assert.notEqual(undefined, data.response);
              assert.notEqual(null, data.response);
              assert.equal(0, data.response.indexOf('{crypt}'));
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    // describe('#hasEntity', () => {
    //   it('should have a hasEntity function', (done) => {
    //     try {
    //       assert.equal(true, typeof a.hasEntity === 'function');
    //       done();
    //     } catch (error) {
    //       log.error(`Test Failure: ${error}`);
    //       done(error);
    //     }
    //   });
    //   it('should find entity', (done) => {
    //     try {
    //       a.hasEntity('template_entity', // 'a9e9c33dc61122760072455df62663d2', (data) => {
    //         try {
    //           assert.equal(true, data[0]);
    //           done();
    //         } catch (err) {
    //           log.error(`Test Failure: ${err}`);
    //           done(err);
    //         }
    //       });
    //     } catch (error) {
    //       log.error(`Adapter Exception: ${error}`);
    //       done(error);
    //     }
    //   }).timeout(attemptTimeout);
    //   it('should not find entity', (done) => {
    //     try {
    //       a.hasEntity('template_entity', 'blah', (data) => {
    //         try {
    //           assert.equal(false, data[0]);
    //           done();
    //         } catch (err) {
    //           log.error(`Test Failure: ${err}`);
    //           done(err);
    //         }
    //       });
    //     } catch (error) {
    //       log.error(`Adapter Exception: ${error}`);
    //       done(error);
    //     }
    //   }).timeout(attemptTimeout);
    // });

    /*
    -----------------------------------------------------------------------
    -----------------------------------------------------------------------
    *** All code above this comment will be replaced during a migration ***
    ******************* DO NOT REMOVE THIS COMMENT BLOCK ******************
    -----------------------------------------------------------------------
    -----------------------------------------------------------------------
    */

    describe('#listStacks - errors', () => {
      it('should have a listStacks function', (done) => {
        try {
          assert.equal(true, typeof a.listStacks === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStacks - errors', () => {
      it('should have a describeStacks function', (done) => {
        try {
          assert.equal(true, typeof a.describeStacks === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#createStack - errors', () => {
      it('should have a createStack function', (done) => {
        try {
          assert.equal(true, typeof a.createStack === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.createStack(null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-createStack', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#updateStack - errors', () => {
      it('should have a updateStack function', (done) => {
        try {
          assert.equal(true, typeof a.updateStack === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.updateStack(null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-updateStack', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#updateTerminationProtection - errors', () => {
      it('should have a updateTerminationProtection function', (done) => {
        try {
          assert.equal(true, typeof a.updateTerminationProtection === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.updateTerminationProtection(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-updateTerminationProtection', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing enableTerminationProtection', (done) => {
        try {
          a.updateTerminationProtection('fake data', null, (data, error) => {
            try {
              const displayE = 'enableTerminationProtection is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-updateTerminationProtection', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#deleteStack - errors', () => {
      it('should have a deleteStack function', (done) => {
        try {
          assert.equal(true, typeof a.deleteStack === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.deleteStack(null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-deleteStack', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listStackInstances - errors', () => {
      it('should have a listStackInstances function', (done) => {
        try {
          assert.equal(true, typeof a.listStackInstances === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.listStackInstances(null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackInstance - errors', () => {
      it('should have a describeStackInstance function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackInstance === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.describeStackInstance(null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackInstance', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackInstanceAccount', (done) => {
        try {
          a.describeStackInstance('fake data', null, null, (data, error) => {
            try {
              const displayE = 'stackInstanceAccount is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackInstance', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackInstanceRegion', (done) => {
        try {
          a.describeStackInstance('fake data', 'fake data', null, (data, error) => {
            try {
              const displayE = 'stackInstanceRegion is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackInstance', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#createStackInstances - errors', () => {
      it('should have a createStackInstances function', (done) => {
        try {
          assert.equal(true, typeof a.createStackInstances === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.createStackInstances(null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-createStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing regions', (done) => {
        try {
          a.createStackInstances('fake data', null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'regions is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-createStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#updateStackInstances - errors', () => {
      it('should have a updateStackInstances function', (done) => {
        try {
          assert.equal(true, typeof a.updateStackInstances === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.updateStackInstances(null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-updateStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing regions', (done) => {
        try {
          a.updateStackInstances('fake data', null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'regions is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-updateStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#deleteStackInstances - errors', () => {
      it('should have a deleteStackInstances function', (done) => {
        try {
          assert.equal(true, typeof a.deleteStackInstances === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.deleteStackInstances(null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-deleteStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing regions', (done) => {
        try {
          a.deleteStackInstances('fake data', null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'regions is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-deleteStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing retainStacks', (done) => {
        try {
          a.deleteStackInstances('fake data', ['fake data'], null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'retainStacks is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-deleteStackInstances', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listStackSets - errors', () => {
      it('should have a listStackSets function', (done) => {
        try {
          assert.equal(true, typeof a.listStackSets === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackSet - errors', () => {
      it('should have a describeStackSet function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.describeStackSet(null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#createStackSet - errors', () => {
      it('should have a createStackSet function', (done) => {
        try {
          assert.equal(true, typeof a.createStackSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.createStackSet(null, null, null, null, null, null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-createStackSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#updateStackSet - errors', () => {
      it('should have a updateStackSet function', (done) => {
        try {
          assert.equal(true, typeof a.updateStackSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.updateStackSet(null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-updateStackSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#deleteStackSet - errors', () => {
      it('should have a deleteStackSet function', (done) => {
        try {
          assert.equal(true, typeof a.deleteStackSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.deleteStackSet(null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-deleteStackSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#cancelUpdateStack - errors', () => {
      it('should have a cancelUpdateStack function', (done) => {
        try {
          assert.equal(true, typeof a.cancelUpdateStack === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.cancelUpdateStack(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-cancelUpdateStack', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listStackSetOperations - errors', () => {
      it('should have a listStackSetOperations function', (done) => {
        try {
          assert.equal(true, typeof a.listStackSetOperations === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.listStackSetOperations(null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listStackSetOperations', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listStackSetOperationResults - errors', () => {
      it('should have a listStackSetOperationResults function', (done) => {
        try {
          assert.equal(true, typeof a.listStackSetOperationResults === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.listStackSetOperationResults(null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listStackSetOperationResults', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing operationId', (done) => {
        try {
          a.listStackSetOperationResults('fake data', null, null, null, (data, error) => {
            try {
              const displayE = 'operationId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listStackSetOperationResults', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackSetOperation - errors', () => {
      it('should have a describeStackSetOperation function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackSetOperation === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.describeStackSetOperation(null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackSetOperation', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing operationId', (done) => {
        try {
          a.describeStackSetOperation('fake data', null, (data, error) => {
            try {
              const displayE = 'operationId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackSetOperation', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#stopStackSetOperation - errors', () => {
      it('should have a stopStackSetOperation function', (done) => {
        try {
          assert.equal(true, typeof a.stopStackSetOperation === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.stopStackSetOperation(null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-stopStackSetOperation', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing operationId', (done) => {
        try {
          a.stopStackSetOperation('fake data', null, (data, error) => {
            try {
              const displayE = 'operationId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-stopStackSetOperation', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackEvents - errors', () => {
      it('should have a describeStackEvents function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackEvents === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listStackResources - errors', () => {
      it('should have a listStackResources function', (done) => {
        try {
          assert.equal(true, typeof a.listStackResources === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.listStackResources(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listStackResources', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackResource - errors', () => {
      it('should have a describeStackResource function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackResource === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.describeStackResource(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackResource', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing logicalResourceId', (done) => {
        try {
          a.describeStackResource('fake data', null, (data, error) => {
            try {
              const displayE = 'logicalResourceId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackResource', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackResources - errors', () => {
      it('should have a describeStackResources function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackResources === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackResourceDrifts - errors', () => {
      it('should have a describeStackResourceDrifts function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackResourceDrifts === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.describeStackResourceDrifts(null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackResourceDrifts', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#detectStackResourceDrift - errors', () => {
      it('should have a detectStackResourceDrift function', (done) => {
        try {
          assert.equal(true, typeof a.detectStackResourceDrift === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.detectStackResourceDrift(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-detectStackResourceDrift', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing logicalResourceId', (done) => {
        try {
          a.detectStackResourceDrift('fake data', null, (data, error) => {
            try {
              const displayE = 'logicalResourceId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-detectStackResourceDrift', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeStackDriftDetectionStatus - errors', () => {
      it('should have a describeStackDriftDetectionStatus function', (done) => {
        try {
          assert.equal(true, typeof a.describeStackDriftDetectionStatus === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackDriftDetectionId', (done) => {
        try {
          a.describeStackDriftDetectionStatus(null, (data, error) => {
            try {
              const displayE = 'stackDriftDetectionId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeStackDriftDetectionStatus', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#detectStackDrift - errors', () => {
      it('should have a detectStackDrift function', (done) => {
        try {
          assert.equal(true, typeof a.detectStackDrift === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.detectStackDrift(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-detectStackDrift', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#detectStackSetDrift - errors', () => {
      it('should have a detectStackSetDrift function', (done) => {
        try {
          assert.equal(true, typeof a.detectStackSetDrift === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackSetName', (done) => {
        try {
          a.detectStackSetDrift(null, null, null, (data, error) => {
            try {
              const displayE = 'stackSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-detectStackSetDrift', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#getStackPolicy - errors', () => {
      it('should have a getStackPolicy function', (done) => {
        try {
          assert.equal(true, typeof a.getStackPolicy === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.getStackPolicy(null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-getStackPolicy', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#setStackPolicy - errors', () => {
      it('should have a setStackPolicy function', (done) => {
        try {
          assert.equal(true, typeof a.setStackPolicy === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.setStackPolicy(null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-setStackPolicy', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#continueUpdateRollback - errors', () => {
      it('should have a continueUpdateRollback function', (done) => {
        try {
          assert.equal(true, typeof a.continueUpdateRollback === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.continueUpdateRollback(null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-continueUpdateRollback', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listChangeSets - errors', () => {
      it('should have a listChangeSets function', (done) => {
        try {
          assert.equal(true, typeof a.listChangeSets === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.listChangeSets(null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listChangeSets', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeChangeSet - errors', () => {
      it('should have a describeChangeSet function', (done) => {
        try {
          assert.equal(true, typeof a.describeChangeSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing changeSetName', (done) => {
        try {
          a.describeChangeSet(null, null, null, (data, error) => {
            try {
              const displayE = 'changeSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeChangeSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#executeChangeSet - errors', () => {
      it('should have a executeChangeSet function', (done) => {
        try {
          assert.equal(true, typeof a.executeChangeSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing changeSetName', (done) => {
        try {
          a.executeChangeSet(null, null, null, (data, error) => {
            try {
              const displayE = 'changeSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-executeChangeSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#createChangeSet - errors', () => {
      it('should have a createChangeSet function', (done) => {
        try {
          assert.equal(true, typeof a.createChangeSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.createChangeSet(null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-createChangeSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing changeSetName', (done) => {
        try {
          a.createChangeSet('fake data', null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'changeSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-createChangeSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#deleteChangeSet - errors', () => {
      it('should have a deleteChangeSet function', (done) => {
        try {
          assert.equal(true, typeof a.deleteChangeSet === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing changeSetName', (done) => {
        try {
          a.deleteChangeSet(null, null, (data, error) => {
            try {
              const displayE = 'changeSetName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-deleteChangeSet', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listTypes - errors', () => {
      it('should have a listTypes function', (done) => {
        try {
          assert.equal(true, typeof a.listTypes === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeType - errors', () => {
      it('should have a describeType function', (done) => {
        try {
          assert.equal(true, typeof a.describeType === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listTypeRegistrations - errors', () => {
      it('should have a listTypeRegistrations function', (done) => {
        try {
          assert.equal(true, typeof a.listTypeRegistrations === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeTypeRegistration - errors', () => {
      it('should have a describeTypeRegistration function', (done) => {
        try {
          assert.equal(true, typeof a.describeTypeRegistration === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing registrationToken', (done) => {
        try {
          a.describeTypeRegistration(null, (data, error) => {
            try {
              const displayE = 'registrationToken is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-describeTypeRegistration', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#registerType - errors', () => {
      it('should have a registerType function', (done) => {
        try {
          assert.equal(true, typeof a.registerType === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing schemaHandlerPackage', (done) => {
        try {
          a.registerType(null, null, null, null, null, null, (data, error) => {
            try {
              const displayE = 'schemaHandlerPackage is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-registerType', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#deregisterType - errors', () => {
      it('should have a deregisterType function', (done) => {
        try {
          assert.equal(true, typeof a.deregisterType === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listTypeVersions - errors', () => {
      it('should have a listTypeVersions function', (done) => {
        try {
          assert.equal(true, typeof a.listTypeVersions === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#setTypeDefaultVersion - errors', () => {
      it('should have a setTypeDefaultVersion function', (done) => {
        try {
          assert.equal(true, typeof a.setTypeDefaultVersion === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listImports - errors', () => {
      it('should have a listImports function', (done) => {
        try {
          assert.equal(true, typeof a.listImports === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing exportName', (done) => {
        try {
          a.listImports(null, null, (data, error) => {
            try {
              const displayE = 'exportName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-listImports', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#listExports - errors', () => {
      it('should have a listExports function', (done) => {
        try {
          assert.equal(true, typeof a.listExports === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#describeAccountLimits - errors', () => {
      it('should have a describeAccountLimits function', (done) => {
        try {
          assert.equal(true, typeof a.describeAccountLimits === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#getTemplate - errors', () => {
      it('should have a getTemplate function', (done) => {
        try {
          assert.equal(true, typeof a.getTemplate === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#getTemplateSummary - errors', () => {
      it('should have a getTemplateSummary function', (done) => {
        try {
          assert.equal(true, typeof a.getTemplateSummary === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#validateTemplate - errors', () => {
      it('should have a validateTemplate function', (done) => {
        try {
          assert.equal(true, typeof a.validateTemplate === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#estimateTemplateCost - errors', () => {
      it('should have a estimateTemplateCost function', (done) => {
        try {
          assert.equal(true, typeof a.estimateTemplateCost === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });

    describe('#signalResource - errors', () => {
      it('should have a signalResource function', (done) => {
        try {
          assert.equal(true, typeof a.signalResource === 'function');
          done();
        } catch (error) {
          log.error(`Test Failure: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing logicalResourceId', (done) => {
        try {
          a.signalResource(null, null, null, null, (data, error) => {
            try {
              const displayE = 'logicalResourceId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-signalResource', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing stackName', (done) => {
        try {
          a.signalResource('fake data', null, null, null, (data, error) => {
            try {
              const displayE = 'stackName is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-signalResource', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing status', (done) => {
        try {
          a.signalResource('fake data', 'fake data', null, null, (data, error) => {
            try {
              const displayE = 'status is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-signalResource', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
      it('should error if - missing uniqueId', (done) => {
        try {
          a.signalResource('fake data', 'fake data', 'fake data', null, (data, error) => {
            try {
              const displayE = 'uniqueId is required';
              runErrorAsserts(data, error, 'AD.300', 'Test-awscloudFormation-adapter-signalResource', displayE);
              done();
            } catch (err) {
              log.error(`Test Failure: ${err}`);
              done(err);
            }
          });
        } catch (error) {
          log.error(`Adapter Exception: ${error}`);
          done(error);
        }
      }).timeout(attemptTimeout);
    });
  });
});
